#include "CameraCommands.h"

#include <P1Camera.hpp>
#include <C_P1CameraCommonProperties.h>
#include <P1CameraExceptions.hpp>

#include <iostream>
#include <fstream>
#include <sstream>
#include <regex>
#include <functional>
#include <atomic>

using std::string;
using std::vector;

#ifdef _MSC_VER
#include <windows.h>
#include <processthreadsapi.h>
#endif

// Anonymous namespace
namespace
{
    std::map<int,string> tetherMap {
        {P1::CameraSdk::TetherConnection::kTetherConnection_Undefined, "-"},
        {P1::CameraSdk::TetherConnection::kTetherConnection_USB,       "USB"},
        {P1::CameraSdk::TetherConnection::kTetherConnection_USB_2,     "USB2"},
        {P1::CameraSdk::TetherConnection::kTetherConnection_USB_3,     "USB3"},
        {P1::CameraSdk::TetherConnection::kTetherConnection_TCPIP,     "IP"}
    };
}

std::atomic<int> CameraCommands::recvCount = 0;

CameraCommands::CameraCommands(Shell& shell)
{
    shell.AddHandler("list",    [this]() { this->PrintList(); }, "Prints a list of visible camera (USB only)");
    shell.AddHandler("open",    [this](vector<string>& tokens) { this->Open(tokens); }, "[IP=X.X.X.X|serialno=<serialno>] [password=<password>] [retries=<num>] [timeout=<secs>] Open an IP or USB camera");
    shell.AddHandler("close",   [this](vector<string>& tokens) { this->Close(tokens); }, "Close the currently opened camera");
    shell.AddHandler("capture", [this](vector<string>& tokens) { this->Capture(tokens); }, "Trigger a capture on the open camera");
    shell.AddHandler("recv",    [this](vector<string>& tokens) { this->EnableReceiving(tokens); }, "<on|off> Enable / Disable receiving captured images.");
    shell.AddHandler("liveview",[this](vector<string>& tokens) { this->SetLiveView(tokens); }, "[on|off] Shows config and Enable/Disable receiving live view images");
    shell.AddHandler("save",    [this](vector<string>& tokens) { this->SetSave(tokens); }, "<on|off> Save received images to CWD");
    shell.AddHandler("hostcapacity", [this](vector<string>& tokens) { this->SetHostCapacity(tokens); }, "<capacity in MB> Tells the camera how much capacity this host has available for images");
    shell.AddHandler("opened",  [this](vector<string>& tokens) { this->List(tokens); }, "List all opened cameras");
    shell.AddHandler("pending", [this](vector<string>& tokens) { this->Pending(tokens); }, "List pending images count, for all cameras");
}

CameraCommands::~CameraCommands()
{
    TerminateImageThread();
    TerminateLiveViewThread();
}


void CameraCommands::PrintList()
{
    auto cams = P1::CameraSdk::Camera::GetAvailableCameras();
    if (cams.size() == 0)
    {
        std::cout << "No cameras are visible! (Remember only Bonjour and USB cameras will appear in the list)\n";
        return;
    }

    for (auto camera : cams)
    {
        if (camera.mConnection == P1::CameraSdk::kTetherConnection_TCPIP)
        {
            std::cout << " * " << camera.mName << " : " << (camera.mSerialNum.empty() ? "" : camera.mSerialNum) << " (" << tetherMap[camera.mConnection] << ": " << camera.mIPv4 << ")" << (camera.mIsOpen ? " OPEN" : "") << std::endl;
        }
        else
        {
            std::cout << " * " << camera.mName << " : " << camera.mSerialNum << " (" << tetherMap[camera.mConnection] << ")" << (camera.mIsOpen ? " OPEN" : "") << std::endl;
        }
    }
}


void CameraCommands::Open(vector<string>& tokens)
{
    // Next field/parameter in tokens to be inspected:
    uint32_t fieldNo = 1;   // "open" was idx=0

    string password;
    {   // Look for "password=<password>" parameter among tokens (no surrounding quotes, and no spaces allowed, here at least)
        std::regex passwordRegex("password=(.*)", std::regex_constants::icase);
        std::smatch m;
        for (int i = fieldNo; i < tokens.size(); i++)
        {
            if (std::regex_search(tokens[i], m, passwordRegex) && (m.size() == 2))  // Exactly two matches, whole string and group with just RHS
            {
                password = m[1].str();
            }
        }
    }

    string serialno;
    {   // Look for "serialno=<serialno>" parameter among tokens (no surrounding quotes, and no spaces allowed, here at least)
        std::regex serialRegex("serialno=(.*)", std::regex_constants::icase);
        std::smatch m;
        for (int i = fieldNo; i < tokens.size(); i++)
        {
            if (std::regex_search(tokens[i], m, serialRegex) && (m.size() == 2))  // Exactly two matches, whole string and group with just RHS
            {
                serialno = m[1].str();
            }
        }
    }

    uint32_t retries = 1;   // At least one...
    {   // Look for "retries=<num>" parameter among tokens
        std::regex retriesRegex("retries=(.*)", std::regex_constants::icase);
        std::smatch m;
        for (int i = fieldNo; i < tokens.size(); i++)
        {
            if (std::regex_search(tokens[i], m, retriesRegex) && (m.size() == 2))  // Exactly two matches, whole string and group with just RHS
            {
                retries = atoi(m[1].str().c_str());
            }
        }
    }

    int32_t timeout = 0;
    {   // Look for "timeout=<uint>" parameter among tokens
        std::regex retriesRegex("timeout=(\\d+)", std::regex_constants::icase);
        std::smatch m;
        for (int i = fieldNo; i < tokens.size(); i++)
        {
            if (std::regex_search(tokens[i], m, retriesRegex) && (m.size() == 2))  // Exactly two matches, whole string and group with just RHS
            {
                timeout = atoi(m[1].str().c_str());

                if (timeout == 0)
                {
                    std::cout << "A timeout of 0, is no system's default timeout!\n";
                }
                else if (timeout < 0)
                {
                    std::cout << "Error: Negative timeouts does not make any sense!\n";
                    return;
                }
            }
        }
    }


    if (Shell::IndexOfArgument("all", tokens) != -1)
    {
        std::cout << "Opening all USB cameras...\n";
        for (auto moniker : P1::CameraSdk::Camera::GetAvailableCameras())
        {
            if (moniker.mConnection < 1 && moniker.mConnection > 3) continue;
            std::cout << "Try open " << moniker.mSerialNum << " (" << moniker.mName << ")\n";

            try
            {
                auto camera = P1::CameraSdk::Camera::OpenUsbCamera(moniker.mSerialNum, password);
                std::pair<const int, P1::CameraSdk::Camera> pair = { camera.GetId(), std::move(camera) };
                mCameras.insert(std::move(pair));
            }
            catch (const P1::CameraSdk::SdkException& e)
            {
                std::cout << "failed to open " << moniker.mName << ". Error: " << e.what() << std::endl;
            }
        }

        return;
    }


    string ipAddress;
    {   // Look for ip=<ip-address>
        std::regex ipRegex("IP=(\\d+\\.\\d+\\.\\d+\\.\\d+)", std::regex_constants::icase);
        std::smatch m;
        for (int i = fieldNo; i < tokens.size(); i++)
        {
            if (std::regex_search(tokens[fieldNo], m, ipRegex) && (m.size() == 2))  // Exactly two matches, whole string and group with just RHS
            {
                ipAddress = m[1].str();
            }
        }
    }

    // Connection
    do
    {
        if (false == ipAddress.empty())
        {   // IP
            std::cout << "Connecting to IP Camera...\n";
            try
            {

                auto camera = P1::CameraSdk::Camera::OpenIpCamera(ipAddress, password, timeout);
                std::pair<const int, P1::CameraSdk::Camera> pair = { camera.GetId(), std::move(camera) };
                mCameras.insert(std::move(pair));
                std::cout << "Opened camera with id: " << pair.first << std::endl;
                return;
            }
            catch (const P1::CameraSdk::SdkException& e)
            {
                std::cout << "Exception caught while opening IP Camera: \"" << e.what() << "\", ErrorCode: " << e.mErrorCode << std::endl;
            }
        }
        else
        {   // USB
            std::cout << "Trying USB...\n";
            try
            {
                auto camera = P1::CameraSdk::Camera::OpenUsbCamera(serialno, password);
                std::pair<const int, P1::CameraSdk::Camera> pair = { camera.GetId(), std::move(camera) };
                mCameras.insert(std::move(pair));
                std::cout << "Opened camera with id: " << pair.first << std::endl;
                return;
            }
            catch (const P1::CameraSdk::SdkException& e)
            {
                std::cout << "Exception caught while opening USB Camera: \"" << e.what() << "\", ErrorCode: " << e.mErrorCode << std::endl;
            }
        }

        retries--;
    } while (retries);
}


void CameraCommands::Close(std::vector<std::string>& tokens)
{
    // Join our threads
    TerminateImageThread();
    TerminateLiveViewThread();

    if (Shell::IndexOfArgument("all", tokens) != -1)
    {
        for (auto& camera : mCameras)
        {
            if (camera.second.IsOpen())
                camera.second.Close();
        }
    }
    else
    {
        GetCameraConnected().Close();
    }

    mCameras.erase(mCameras.begin());
}

void cameraTriggerWithCount(P1::CameraSdk::Camera& camera, uint32_t triggerCount, uint32_t delayMs)
{
    for (uint32_t i = 0; i < triggerCount; i++)
    {
        camera.TriggerCapture();
        if (delayMs > 0)
            std::this_thread::sleep_for(std::chrono::milliseconds(delayMs));
    }
}

void CameraCommands::Capture(std::vector<std::string>& tokens)
{
    bool captureOnAll = false;
    uint32_t count = 1;
    uint32_t intervalMs = 1000;
    captureOnAll = Shell::IndexOfArgument("all", tokens, true) != -1;

    if (Shell::IndexOfArgument("-c", tokens) != -1)
    {
        int index = Shell::IndexOfArgument("-c", tokens);
        if (tokens.size() > index+1)
        {
            count = std::atoi(tokens[index+1].c_str());
            std::cout << "Triggers " << count << " image" << (count==1?"\n":"s\n");
        }
    }

    if (Shell::IndexOfArgument("-i", tokens) != -1)
    {
        int index = Shell::IndexOfArgument("-i", tokens);
        if (tokens.size() > index + 1)
        {
            intervalMs = std::atoi(tokens[index + 1].c_str());
            std::cout << "interval is " << intervalMs << "ms\n";
        }
    }

    try
    {
        if (captureOnAll == false)
        {
            for (uint32_t i = 0; i < count; i++)
            {
                GetCameraConnected().TriggerCapture();
                std::cout << ".";
                std::this_thread::sleep_for(std::chrono::milliseconds(intervalMs));
            }
        }
        else
        {
            std::cout << "Trigger on " << mCameras.size() << " camera" << (mCameras.size() != 1 ? "s\n":"\n");
            for (uint32_t i = 0; i < count; i++)
            {
                for (auto& camera : mCameras)
                {
                    camera.second.TriggerCapture();
                }
                std::cout << ".";
                std::this_thread::sleep_for(std::chrono::milliseconds(intervalMs));
            }
        }

        std::cout << std::endl;
    }
    catch (P1::CameraSdk::SdkException e)
    {
        std::cout << "Exception caught while capturing: \"" << e.what() << "\", ErrorCode: " << e.mErrorCode << std::endl;
    }
}


void CameraCommands::EnableReceiving(vector<string>& tokens)
{
    if (tokens.size() < 2) {
        std::cout << "Missing on|off argument. Usage: " << tokens[0] << "on|off|count [-reset]\n";
        return;
    }

    try
    {
        if (Shell::BeginsWith("on", tokens[1], true))
        {
            if (mImageThreadRunnable == true) return;

            for (auto& camera : mCameras)
            {
                camera.second.EnableImageReceiving(true);
            }

            mImageThreadRunnable = true;
            for (auto& camera : mCameras)
            {
                mImageThreads.insert({ camera.first, std::unique_ptr<std::thread>(new std::thread([this, &camera]() {
#ifdef _MSC_VER
                    SetThreadDescription(GetCurrentThread(), L"Camera Receive Image");
#endif
                    this->ImageThreadMain(camera.second);
                })) });
            }
        }
        else if (Shell::BeginsWith("off", tokens[1], true))
        {
            if (mImageThreadRunnable == false) return;

            mImageThreadRunnable = false;

            for (auto& camera : mCameras)
            {
                camera.second.EnableImageReceiving(false);
                mImageThreads[camera.first]->join();
                mImageThreads.erase(camera.first);
            }
        }
        else if (Shell::BeginsWith("count", tokens[1], true))
        {
            if (Shell::IndexOfArgument("-reset", tokens, true) != -1)
            {
                CameraCommands::recvCount = 0;
            }

            std::cout << "Receive count: " << CameraCommands::recvCount << std::endl;
        }
        else
        {
            std::cout << "Unknown parameter: " << tokens[1] << "Usage: " << tokens[0] << " on|off|count [-reset]\n";
        }
    }
    catch (P1::CameraSdk::SdkException e)
    {
        std::cout << "Exception caught handling recv command: \"" << e.what() << "\", ErrorCode: " << e.mErrorCode << std::endl;
    }
}


void CameraCommands::SetHostCapacity(vector<string>& tokens)
{
    if (tokens.size() < 2) {
        std::cout << "Missing <capacity in MB> argument. Usage: " << tokens[0] << "<capacity in MB>\n";
        return;
    }

    int capacity = atoi(tokens[1].c_str());

    if (capacity < 0)
    {
        std::cout << "Illegal capacity given. Please provide a non-negative number.\n";
        return;
    }

    try
    {
        GetCameraConnected().SetHostStorageCapacity((uint32_t)capacity);
        std::cout << "Host storage capacity updated to: " << (uint32_t)capacity << " MB\n";
    }
    catch (P1::CameraSdk::SdkException e)
    {
        std::cout << "Exception caught handling command: \"" << e.what() << "\", ErrorCode: " << e.mErrorCode << std::endl;
    }
}


void CameraCommands::SetLiveView(vector<string>& tokens)
{
    if (tokens.size() == 1)
    {
        for (auto& cameraPair : mCameras)
        {
            std::cout << "[" << cameraPair.first << "] " << cameraPair.second.GetLiveViewConfiguration().ToString() << "\n" << std::endl;
        }

        return;
    }

    try
    {
        if (Shell::BeginsWith("on", tokens[1], true))
        {
            if (mLiveViewThreadRunnable == true) return;

            for (auto& camera : mCameras)
            {
                camera.second.EnableLiveView();
            }

            mLiveViewThreadRunnable = true;
            for (auto& camera : mCameras)
            {
                mLiveViewThreads.insert({ camera.first, std::unique_ptr<std::thread>(new std::thread([this, &camera]() {
#ifdef _MSC_VER
                    SetThreadDescription(GetCurrentThread(), L"Camera LiveView Receiver");
#endif
                    this->LiveViewThreadMain(camera.second);
                })) });
            }

        }
        else if (Shell::BeginsWith("off", tokens[1], true))
        {
            if (mLiveViewThreadRunnable == false) return;

            mLiveViewThreadRunnable = false;

            for (auto& camera : mCameras)
            {
                camera.second.DisableLiveView();
                mLiveViewThreads[camera.first]->join();
                mLiveViewThreads.erase(camera.first);
            }
        }
        else
        {
            std::cout << "Unknown parameter: " << tokens[1] << "Usage: " << tokens[0] << " on|off\n";
        }
    }
    catch (P1::CameraSdk::SdkException e)
    {
        std::cout << "Exception caught handling recv command: \"" << e.what() << "\", ErrorCode: " << e.mErrorCode << std::endl;
    }
}

void CameraCommands::SetSave(vector<string>& tokens)
{
    if (tokens.size() < 2) {
        std::cout << "Missing on|off argument. Usage: " << tokens[0] << "on|off\n";
        return;
    }

    if (Shell::BeginsWith("on", tokens[1], true))
    {
        saveImages = true;
    }
    else if (Shell::BeginsWith("off", tokens[1], true))
    {
        saveImages = false;
    }
    else
    {
        std::cout << "Unknown parameter: " << tokens[1] << "Usage: " << tokens[0] << " on|off\n";
    }
}

void CameraCommands::Pending(std::vector<std::string>& tokens)
{
    for (auto& pair : mCameras)
    {
        auto pending = pair.second.GetProperty(1200);
        auto ready = pair.second.GetProperty(1201);
        std::cout << " * Camera " << pair.first << ": Pending images: " << pending.ToString() << ", ready to capture: " << ready.ToString() << std::endl;
    }
}

void CameraCommands::TerminateImageThread()
{
    if (mImageThreads.size() > 0)
    {
        mImageThreadRunnable = false;

        for (auto& camPair : mCameras)
        {
            try
            {
                camPair.second.EnableImageReceiving(false);
                std::cout << "Waiting for receive thread for camera " << camPair.first << "to join...\n";
                mImageThreads[camPair.first]->join();
            }
            catch (const P1::CameraSdk::SdkException& e)
            {
                std::cout << "Exception while shutting down image-thread:" << std::endl;
                std::cout << "SDK Error: " << e.mErrorCode << ": Msg: " << e.what() << std::endl;
            }

        }
    }
}

void CameraCommands::ImageThreadMain(P1::CameraSdk::Camera& camera)
{
    std::cout << "Image Thread woke!\n";

    while(mImageThreadRunnable && camera.IsOpen())
    {
        try
        {
            auto fileRef = camera.WaitForImage();
            CameraCommands::recvCount++;
            std::cout << "Camera Id: "<< camera.GetId() << " : Got IIQ file: " << fileRef.filename << " (#" << recvCount << ")\n";

            if (saveImages)
            {
                std::ofstream imageFile(fileRef.filename);
                imageFile.write(reinterpret_cast<const char*>(fileRef.data.get()), fileRef.size);
            }
        }
        catch (const P1::CameraSdk::SdkException& e)
        {
            if (e.mErrorCode == P1::CameraSdk::kErrorImageReceiverTimeout)
            {
                continue;
            }
            if (e.mErrorCode == P1::CameraSdk::kErrorImageReceiverAborted)
            {
                std::cout << "WaitForImage aborted (CameraId: " << camera.GetId() << ")\n";
            }
            else
            {
                std::cerr << "SDK Error: " << e.mErrorCode << ": Msg: " << e.what() << '\n';
                break;
            }
        }
    }

    std::cout << "Image Thread terminate\n";
    mImageThreadRunnable = false;
}

void CameraCommands::LiveViewThreadMain(P1::CameraSdk::Camera& camera)
{
    std::cout << "LiveView Thread woke for camera: " << camera.GetId() << "\n";

    static uint8_t buffer[4*1024*1024];
    auto filename = std::string("liveview_") + std::to_string(camera.GetId()) + std::string(".bin");

    while(mLiveViewThreadRunnable && camera.IsOpen())
    {
        try
        {
            auto meta = camera.WaitForLiveView();
            std::cout << "[" << camera.GetId() << "] " << meta.ToString() << std::endl;
            std::ofstream lvfile(filename, std::ios::trunc);
            lvfile.write(reinterpret_cast<const char*>(meta.imageData.get()), meta.imageSize);
        }
        catch(const P1::CameraSdk::SdkException& e)
        {
            if (e.mErrorCode == P1::CameraSdk::kErrorImageReceiverAborted)
            {
                std::cout << "WaitForLiveView aborted for camera " << camera.GetId() << "\n";
            }
            else
            {
                std::cerr << "Camera: " << camera.GetId() << ", SDK Error: "  << e.mErrorCode << ": Msg: " << e.what() << '\n';
                break;
            }
        }
    }
}

void CameraCommands::TerminateLiveViewThread()
{
    if (mLiveViewThreads.size() > 0)
    {
        mLiveViewThreadRunnable = false;

        for (auto& camPair : mCameras)
        {
            try
            {
                camPair.second.DisableLiveView();
                std::cout << "Waiting for LiveView thread for camera " << camPair.first << "to join...\n";
                mLiveViewThreads[camPair.first]->join();
            }
            catch (const P1::CameraSdk::SdkException& e)
            {
                std::cout << "Exception while shutting down image-thread:" << std::endl;
                std::cout << "SDK Error: " << e.mErrorCode << ": Msg: " << e.what() << std::endl;
            }

        }
    }
}

void CameraCommands::List(std::vector<std::string>& tokens)
{
    std::cout << "Opened cameras:\n";
    for (const std::pair<const int, P1::CameraSdk::Camera>& pair : mCameras)
    {
        if (pair.second.IsOpen() == false) continue;

        auto model = pair.second.GetProperty(2);
        auto serial = pair.second.GetProperty(3);
        std::cout << " [" << pair.first << "] : " << model.StringValue() << " " << serial.StringValue() << std::endl;
    }
}
