#include "NotificationCommands.h"

#include <vector>
#include <iostream>
#include <sstream>
#include <chrono>
#include <ctime>
#include <string>

#include <P1CameraListener.hpp>
#include <P1CameraExceptions.hpp>

using std::string;
using std::vector;
using std::cout;

using namespace P1::CameraSdk;

// Anonymous namespace
namespace
{
    //const SdkEventId kSdkEvents[] = { kSdkEventCameraAdded, kSdkEventCameraRemoved };

    struct EventAndName
    {
        uint32_t eventId;
        std::string name;
    };

    const EventAndName kSdkEventNameMapping[] = {
        {kSdkEventCameraFound,   "SdkCameraFound"},
        {kSdkEventCameraRemoved, "SdkCameraRemoved"},
        {kSdkEventCameraOpened,  "SdkCameraOpened"},
        {kSdkEventCameraClosed,  "SdkCameraClosed"}
    };

    const EventAndName kCameraEventNameMapping[] = {
        {kCameraEventImageTransferBegin,    "CameraImageTransferBegin"},
        {kCameraEventImageTransferProgress, "CameraImageTransferProgress"},
        {kCameraEventImageReady,            "CameraImageReady"},
        {kCameraEventDisconnected,          "CameraDisconnected"},
        {kCameraEventLiveViewStarted,       "CameraLiveViewStarted"},
        {kCameraEventLiveViewStopped,       "CameraLiveViewStopped"},
        {kCameraEventNewLiveViewImage,      "CameraNewLiveViewImage"},
        {kCameraEventPropertyAdded,         "CameraPropertyAdded"},
        {kCameraEventPropertyRemoved,       "CameraPropertyRemoved"}
    };
}

NotificationCommands::NotificationCommands(Shell& shell, ICameraProvider& cameraProvider) :
    mCameraProvider(cameraProvider)
{
    shell.AddHandler("notif",   [this](vector<string>& tokens) { this->HandleNotificationCommand(tokens); }, "Enable/Disable notifications");
}

void NotificationCommands::HandleNotificationCommand(vector<string>& tokens)
{
    if (tokens.size() == 0)
    {
        std::cout << "Critical command error - no arguments passed\n";
        return;
    }
    else if (tokens.size() == 1)
    {
        std::cout << "Missing parameters. Usage:\n";
        std::cout << tokens[0] << " sdk enable|disable\n";
        std::cout << tokens[0] << " camera enable|disable (On all opened cameras)\n";
        std::cout << tokens[0] << " property <id> enable|disable\n";
        std::cout << tokens[0] << " propSpec <id> enable|disable\n";
        std::cout << tokens[0] << " catchAll enable|disable\n";
        return;
    }

    // Lazy construction of the listener-instance:
    if (mListener == nullptr)
    {
        mListener = new P1::CameraSdk::Listener();
    }

    if (mNotificationThread == nullptr)
    {
        mNotificationThread = new std::thread(&NotificationCommands::NotificationThreadMain, this);
    }

    try
    {
        uint32_t fieldNo = 1;
        if (Shell::BeginsWith("sdk", tokens[fieldNo], true))
        {
            HandleSdkNotifCommand(tokens, fieldNo + 1);
        }
        else if (Shell::BeginsWith("camera", tokens[fieldNo], true))
        {
            HandleCameraNotifCommand(tokens, fieldNo + 1);
        }
        else if (Shell::BeginsWith("property", tokens[fieldNo], true))
        {
            HandlePropertyNotifCommand(tokens, fieldNo + 1, false);
        }
        else if (Shell::BeginsWith("propSpec", tokens[fieldNo], true))
        {
            HandlePropertyNotifCommand(tokens, fieldNo + 1, true);
        }
        else if (Shell::BeginsWith("catchAll", tokens[fieldNo], true))
        {
            HandleCatchAllCommand(tokens, fieldNo + 1);
        }
        else
        {
            cout << "Unhandled parameter - aborting command\n";
        }
    }
    catch (P1::CameraSdk::SdkException e)
    {
        std::cout << "Exception caught handling notification command: \"" << e.what() << "\", ErrorCode: " << e.mErrorCode << std::endl;
    }
}

void NotificationCommands::HandleSdkNotifCommand(std::vector<std::string>& tokens, uint32_t fieldNo)
{
    auto remainingNoParameters = tokens.size() - fieldNo;
    if (remainingNoParameters < 1)
    {
        cout << "Missing enable|disable parameter...\n";
        return;
    }

    bool doEnable = true;

    if (Shell::BeginsWith("enable", tokens[fieldNo], true))
        doEnable = true;
    else if (Shell::BeginsWith("disable", tokens[fieldNo], true))
        doEnable = false;
    else
    {
        cout << "Unhandled parameter: " << tokens[fieldNo] << std::endl;
        return;
    }

    for (int i = 0; i < sizeof(kSdkEventNameMapping) / sizeof(kSdkEventNameMapping[0]); i++)
    {
        if (doEnable)
            mListener->EnableNotification((SdkEventId)kSdkEventNameMapping[i].eventId);
        else
            mListener->DisableNotification((SdkEventId)kSdkEventNameMapping[i].eventId);
    }
}

void NotificationCommands::setCameraEvent(bool doEnable, P1::CameraSdk::CameraEventId eventId)
{
    for (auto& pair : mCameraProvider.AllCameras())
    {
        if (doEnable)
            mListener->EnableNotification(pair.second, eventId);
        else
            mListener->DisableNotification(pair.second, eventId);
    }
}

void NotificationCommands::HandleCameraNotifCommand(std::vector<std::string>& tokens, uint32_t fieldNo)
{
    auto& camera = mCameraProvider.GetCameraConnected();
    if (false == camera.IsOpen())
    {
        cout << "No open camera found. Aborting.\n";
        return;
    }

    auto remainingNoParameters = tokens.size() - fieldNo;
    if (remainingNoParameters < 1)
    {
        cout << "Missing enable|disable parameter...\n";
        return;
    }

    bool doEnable = true;

    if (Shell::BeginsWith("enable", tokens[fieldNo], true))
        doEnable = true;
    else if (Shell::BeginsWith("disable", tokens[fieldNo], true))
        doEnable = false;
    else
    {
        cout << "Unhandled parameter: " << tokens[fieldNo] << std::endl;
        return;
    }

    for (int i = 0; i < sizeof(kCameraEventNameMapping) / sizeof(kCameraEventNameMapping[0]); i++)
    {
        setCameraEvent(doEnable, static_cast<P1::CameraSdk::CameraEventId>(kCameraEventNameMapping[i].eventId));
    }
}

void NotificationCommands::HandlePropertyNotifCommand(std::vector<std::string>& tokens, uint32_t fieldNo, bool isSpecNotif)
{
    auto& camera = mCameraProvider.GetCameraConnected();
    if (false == camera.IsOpen())
    {
        cout << "No open camera found. Aborting.\n";
        return;
    }

    auto remainingNoParameters = tokens.size() - fieldNo;
    if (remainingNoParameters < 2)
    {
        cout << "Missing either <id> and/or enable|disable parameter...\n";
        return;
    }

    int propertyId;
    std::istringstream ss(tokens[fieldNo]);
    ss >> propertyId;
    fieldNo++;  // Consume propertyId

    bool doEnable = true;
    if (Shell::BeginsWith("enable", tokens[fieldNo], true))
        doEnable = true;
    else if (Shell::BeginsWith("disable", tokens[fieldNo], true))
        doEnable = false;
    else
    {
        cout << "Unhandled parameter: " << tokens[fieldNo] << std::endl;
        return;
    }

    bool allCameras = false;
    if (remainingNoParameters > 2 && tokens[fieldNo + 1].compare("all") == 0)
    {
        allCameras = true;
    }

    if (doEnable)
    {
        if (allCameras)
        {
            for (auto& cam : mCameraProvider.AllCameras())
            {
                mListener->EnableNotification(cam.second, propertyId, isSpecNotif);
            }
        }
        else
        {
            mListener->EnableNotification(camera, propertyId, isSpecNotif);
        }
    }
    else
    {
        if (allCameras)
        {
            for (auto& cam : mCameraProvider.AllCameras())
            {
                mListener->DisableNotification(cam.second, propertyId, isSpecNotif);
            }
        }
        else
        {
            mListener->DisableNotification(camera, propertyId, isSpecNotif);
        }
    }
}

void NotificationCommands::HandleCatchAllCommand(std::vector<std::string>& tokens, uint32_t fieldNo)
{
    auto remainingNoParameters = tokens.size() - fieldNo;
    if (remainingNoParameters < 1)
    {
        cout << "Missing enable|disable parameter...\n";
        return;
    }

    bool doEnable = true;
    if (Shell::BeginsWith("enable", tokens[fieldNo], true))
        doEnable = true;
    else if (Shell::BeginsWith("disable", tokens[fieldNo], true))
        doEnable = false;
    else
    {
        cout << "Unhandled parameter: " << tokens[fieldNo] << ". Must be either enable or disable\n";
        return;
    }

    if (doEnable)
    {
        cout << "Enabling catch-all notifications... (implicit: across all cameras)\n";
        mListener->EnableCatchAllNotifications();
    }
    else
    {
        cout << "Disabling catch-all notifications...\n";
        mListener->DisableCatchAllNotifications();
    }
}

NotificationCommands::~NotificationCommands()
{
    if (mListener != nullptr)
    {   // We clean up the listener:
        auto listener = mListener;
        mListener = nullptr;

        // cout << "Waking up notification thread(s)\n";
        listener->WakeUpWaitingThread();
        delete listener;
    }

    if (mNotificationThread != nullptr)
    {
        auto tmpThread = mNotificationThread;
        mNotificationThread = nullptr;

        // cout << "Joining our notif-thread...\n";
        tmpThread->join();
        // cout << "Joined!\n";
        delete tmpThread;
    }
}

void NotificationCommands::NotificationThreadMain()
{
    cout << "NotifThread: Started!\n";

    while (mListener != nullptr)
    {   // We stop the thread, when the listener is null'ed
        try
        {
            const uint32_t kTimeoutMs = 0;
            P1::CameraSdk::NotificationEvent notifEvent;
            bool success = mListener->WaitForNotification(notifEvent, kTimeoutMs);

            if (false == success)   // Timeout
                continue;

            cout << notifEvent.ToString() << std::endl;
        }
        catch (P1::CameraSdk::SdkException e)
        {
            cout << "Exception caught while waiting for notification: \"" << e.what() << "\", ErrorCode: " << e.mErrorCode << std::endl;
        }
    }

    cout << "NotifThread: Stopping\n";
}
