#include "PropertyCommands.h"

#include <vector>
#include <iostream>
#include <sstream>
#include <string>

#include <P1CameraExceptions.hpp>

using std::string;
using std::vector;
using std::cout;

using namespace P1::CameraSdk;

PropertyCommands::PropertyCommands(Shell& shell, ICameraProvider& cameraProvider) :
    mCameraProvider(cameraProvider)
{
    shell.AddHandler("property",   [this](vector<string>& tokens) { this->HandlePropertyCommand(tokens); }, "Control of properties");
}


PropertyCommands::~PropertyCommands()
{
}


void PropertyCommands::HandlePropertyCommand(vector<string>& tokens)
{
    if (tokens.size() == 0)
    {
        std::cout << "Critical command error - no arguments passed\n";
        return;
    }
    else if (tokens.size() == 1)
    {
        std::cout << "Missing parameters. Usage:\n";
        std::cout << tokens[0] << " list - List all properties\n";
        std::cout << tokens[0] << " set <id> <value> - Set a property's value\n";
        std::cout << tokens[0] << " get <id> - Show details for a property\n";
        return;
    }

    try
    {
        uint32_t fieldNo = 1;
        if (Shell::BeginsWith("list", tokens[fieldNo], true))
        {
            HandleListCommand(tokens, fieldNo + 1);
        }
        else if (Shell::BeginsWith("set", tokens[fieldNo], true))
        {
            HandleSetCommand(tokens, fieldNo + 1);
        }
        else if (Shell::BeginsWith("get", tokens[fieldNo], true))
        {
            HandleGetCommand(tokens, fieldNo + 1);
        }
        else
        {
            cout << "Unhandled parameter - aborting command\n";
        }
    }
    catch (P1::CameraSdk::SdkException e)
    {
        std::cout << "Exception caught handling property command: \"" << e.what() << "\", ErrorCode: " << e.mErrorCode << std::endl;
    }
}


void PropertyCommands::PrintPropertyValue(const P1::CameraSdk::PropertyValue& value, const std::string name, const int id) const
{
    std::cout << id << "\t" << name << "\t" << value.ToString() << std::endl;
}


void PropertyCommands::HandleListCommand(std::vector<std::string>& tokens, uint32_t fieldNo)
{
    auto& camera = mCameraProvider.GetCameraConnected();
    if (false == camera.IsOpen())
    {
        cout << "No open camera found. Aborting.\n";
        return;
    }

    auto ids = camera.GetAllPropertyIds();

    for (auto id : ids)
    {
        auto spec = camera.GetPropertySpec(id);
        PrintPropertyValue(spec.mValue, spec.mName, id);
    }
}


void PropertyCommands::HandleGetCommand(std::vector<std::string>& tokens, uint32_t fieldNo)
{
    auto& camera = mCameraProvider.GetCameraConnected();
    if (false == camera.IsOpen())
    {
        cout << "No open camera found. Aborting.\n";
        return;
    }

    auto remainingNoParameters = tokens.size() - fieldNo;
    if (remainingNoParameters < 1)
    {
        cout << "Missing <id> parameter.\n";
        return;
    }

    int propertyId;
    std::istringstream propertyIdSS(tokens[fieldNo]);
    propertyIdSS >> propertyId;

    try
    {
        using namespace P1::CameraSdk;

        auto property = camera.GetPropertySpec(propertyId);
        std::cout << property.ToString();

        // Print value list if present
        if (property.mValueList.size() > 0)
        {
            std::cout << " {\n";
            int count = 0;
            for (auto iter = property.mValueList.begin(); iter != property.mValueList.end(); iter++, count++)
            {
                std::cout << "\t* " << iter->ToString() << " ( " << iter->StringValue() << " )\n";
            }
            std::cout << " }";
        }
        std::cout << std::endl;
    }
    catch (P1::CameraSdk::SdkException e)
    {
        std::cout << "Exception caught while getting property: \"" << e.what() << "\", ErrorCode: " << e.mErrorCode << std::endl;
    }

}


void PropertyCommands::HandleSetCommand(std::vector<std::string>& tokens, uint32_t fieldNo)
{
    auto& camera = mCameraProvider.GetCameraConnected();
    if (false == camera.IsOpen())
    {
        cout << "No open camera found. Aborting.\n";
        return;
    }

    auto remainingNoParameters = tokens.size() - fieldNo;
    if (remainingNoParameters < 2)
    {
        cout << "Missing either <id> and/or <value> parameter...\n";
        return;
    }

    int propertyId;

    std::stringstream propertyIdSS;
    if (Shell::BeginsWith(tokens[(size_t)fieldNo], "0x", true) || Shell::BeginsWith(tokens[(size_t)fieldNo], "x", true))
        propertyIdSS << std::hex;
    propertyIdSS << tokens[(size_t)fieldNo];
    propertyIdSS >> propertyId;

    try
    {
        using namespace P1::CameraSdk;
        auto property = camera.GetProperty(propertyId);

        switch (property.mType)
        {
            case kPropertyTypeBool:
            case kPropertyTypeInt32:
            case kPropertyTypeUInt32:
            case kPropertyTypeInt64:
            case kPropertyTypeUInt64:
            case kPropertyTypeTime:
            case kPropertyTypeEnum:
            {
                std::stringstream ss;
                if (Shell::BeginsWith(tokens[(size_t)fieldNo + 1], "0x", true) || Shell::BeginsWith(tokens[(size_t)fieldNo + 1], "x", true))
                    ss << std::hex;
                ss << tokens[(size_t)fieldNo + 1];
                ss >> property.mInt;
                break;
            }

            case kPropertyTypeFloat64:
            {
                std::stringstream ss;
                ss << tokens[(size_t)fieldNo + 1];
                ss >> property.mDouble;
                break;
            }

            case kPropertyTypeString:
                property.mString = tokens[(size_t)fieldNo + 1];
                break;
            case kPropertyTypeArea:
            {
                std::stringstream ss;
                ss << tokens[(size_t)fieldNo+1] << " ";
                ss << tokens[(size_t)fieldNo+2];
                
                ss >> property.mArea.width;
                ss >> property.mArea.height;
                break;
            }
            case kPropertyTypeAreaFloat:
            {
                std::stringstream ss;
                ss << tokens[(size_t)fieldNo+1] << " ";
                ss << tokens[(size_t)fieldNo+2];
                
                ss >> property.mAreaFloat.width;
                ss >> property.mAreaFloat.height;
                break;
            }
            case kPropertyTypeRange:
            {
                std::stringstream ss;
                ss << tokens[(size_t)fieldNo+1] << " ";
                ss << tokens[(size_t)fieldNo+2];
                
                ss >> property.mRange.min;
                ss >> property.mRange.max;
                break;
            }
            case kPropertyTypeRangeFloat:
            {
                std::stringstream ss;
                ss << tokens[(size_t)fieldNo+1] << " ";
                ss << tokens[(size_t)fieldNo+2];
                
                ss >> property.mRangeFloat.min;
                ss >> property.mRangeFloat.max;
                break;
            }
            case kPropertyTypeColorARGBFloat:
            {
                std::stringstream ss;
                ss << tokens[(size_t)fieldNo+1] << " ";
                ss << tokens[(size_t)fieldNo+2] << " ";
                ss << tokens[(size_t)fieldNo+3] << " ";
                ss << tokens[(size_t)fieldNo+4];
                
                ss >> property.mColor.alpha;
                ss >> property.mColor.red;
                ss >> property.mColor.green;
                ss >> property.mColor.blue;
                break;
            }
            default:
                std::cout << "Sorry! Command line parser can not handle property type" << property.mType << std::endl;
                return;
        }

        camera.SetProperty(propertyId, property);
    }
    catch (P1::CameraSdk::SdkException e)
    {
        std::cout << "Exception caught while setting property: \"" << e.what() << "\", ErrorCode: " << e.mErrorCode << std::endl;
    }
}
