#include "Shell.h"
#include <iostream>
#include <sstream>
#include <iterator>
#include <stdexcept>
#include <iterator>
#include <algorithm>    // For std::transform
#include <cstring> // For strncmp
#include <iomanip>  // For setw

using std::string;
using std::vector;

Shell::Shell(const string& cmdPrefix) :
    prefix(cmdPrefix),
    currentLine(""),
    handlers(),
    shouldRun(true)
{
}

void Shell::ThrowIfExists(const string command)
{
    if (handlers.find(command) != handlers.end())
    {
        auto msg = string("The Shell command ") + command + string(" was added two times!");
        throw std::logic_error(msg);
    }
}

void Shell::AddHandler(const string command, Shell::HandlerFunction handler, const string helpText)
{
    string lowerCaseCommand = Shell::ToLower(command);
    ThrowIfExists(lowerCaseCommand);
    handlers.insert({ lowerCaseCommand, Handler { helpText, handler, std::function<void()>() } });
}

void Shell::AddHandler(const string command, Shell::SimpleHandlerFunction handler, const string helpText)
{
    string lowerCaseCommand = Shell::ToLower(command);
    ThrowIfExists(lowerCaseCommand);
    handlers.insert({ lowerCaseCommand, Handler { helpText, std::function<void(vector<string>)>(), handler } });
}

void Shell::Run()
{
    while (shouldRun)
    {
        std::cout << prefix;
        std::getline(std::cin, currentLine);

        ExecLine(currentLine);
    }
}

void Shell::ExecLine(string line)
{
    //Tokenize
    std::istringstream iss(line);
	std::vector<string> tokens { std::istream_iterator<string>{iss}, std::istream_iterator<string>{} };

    if (handlers.size() == 0)
    {
        std::cout << "No command handlers installed!\n";
        return;
    }

    if (tokens.size() == 0)
    {
        return;
    }

    // We lower-case the command to ignore entered case:
    string enteredCommand = Shell::ToLower(tokens[0]);
    string useThisCommand = enteredCommand;

    // --- BEST MATCH SEARCH
    if (handlers.find(enteredCommand) == handlers.end())
    {   // No single match - let's find best match
        uint32_t numMatches = 0;

        for (auto handler : handlers)
        {
            if (Shell::BeginsWith(handler.first.c_str(), enteredCommand.c_str(), true))
            {   // We have a match...
                useThisCommand = handler.first;
                numMatches++;
            }
        }

        if (numMatches > 1)
        {   // Too many matches
            std::cout << "Ambiguous command entered - " << numMatches << " matches\n";
            return;
        }

        // Nothing - just continue below
    }

    // --- EXECUTION
    if (handlers.find(useThisCommand) != handlers.end())
    {
        if      (handlers[useThisCommand].argFunction)    { handlers[useThisCommand].argFunction(tokens);  }
        else if (handlers[useThisCommand].simpleFunction) { handlers[useThisCommand].simpleFunction();     }
    	else
	    {
	        throw std::logic_error("Missing handler function for command!");
	    }
    }
    else
    {
        std::cout << "No such command: " << tokens[0] << "!" << std::endl;
    }
}

void Shell::SignalStop()
{
    shouldRun = false;
}

void Shell::HelpHandler()
{
    std::cout << "Available commands:\n";
    for (auto pair : this->handlers)
    {
        std::cout << " " << std::setw(10) << std::left << pair.first << pair.second.description << std::endl;;
    }
}

std::string Shell::ToLower(const std::string s)
{
    string lowerCasedString = s;
    transform(lowerCasedString.begin(), lowerCasedString.end(), lowerCasedString.begin(),
        [](unsigned char c) { return std::tolower(c); });

    return lowerCasedString;
}

bool Shell::BeginsWith(const std::string s, const std::string beginsWith, bool ignoreCase)
{
    if (ignoreCase)
        return strncmp(ToLower(s).c_str(), ToLower(beginsWith).c_str(), beginsWith.size()) == 0;
    else
        return strncmp(s.c_str(), beginsWith.c_str(), beginsWith.size()) == 0;
}

int Shell::IndexOfArgument(const std::string argument, const std::vector<std::string> tokens, bool prefixIsMatch)
{
    for (uint32_t index = 0; index < tokens.size(); index++)
    {
        if (prefixIsMatch)
        {
            if (BeginsWith(argument, tokens[index], true))
            {
                return index;
            }
        }
        else
        {
            if (argument.compare(tokens[index]) == 0)
            {
                return index;
            }
        }
    }

    return -1;
}